@@ -93,6 +93,7 @@ impl Id {
 }
 
 /// A trait indicating that a key has parameters.
+#[allow(clippy::missing_safety_doc)]
 pub unsafe trait HasParams {}
 
 unsafe impl HasParams for Params {}
@@ -100,6 +101,7 @@ unsafe impl HasParams for Params {}
 unsafe impl<T> HasParams for T where T: HasPublic {}
 
 /// A trait indicating that a key has public components.
+#[allow(clippy::missing_safety_doc)]
 pub unsafe trait HasPublic {}
 
 unsafe impl HasPublic for Public {}
@@ -107,6 +109,7 @@ unsafe impl HasPublic for Public {}
 unsafe impl<T> HasPublic for T where T: HasPrivate {}
 
 /// A trait indicating that a key has private components.
+#[allow(clippy::missing_safety_doc)]
 pub unsafe trait HasPrivate {}
 
 unsafe impl HasPrivate for Private {}
@@ -117,7 +120,7 @@ generic_foreign_type_and_impl_send_sync! {
 
     /// A public or private key.
     pub struct PKey<T>;
-    /// Reference to `PKey`.
+    /// Reference to [`PKey`].
     pub struct PKeyRef<T>;
 }
 
@@ -276,6 +279,24 @@ where
         private_key_to_der,
         ffi::i2d_PrivateKey
     }
+
+    // This isn't actually PEM output, but `i2d_PKCS8PrivateKey_bio` is documented to be
+    // "identical to the corresponding PEM function", and it's declared in pem.h.
+    private_key_to_pem! {
+        /// Serializes the private key to a DER-encoded PKCS#8 PrivateKeyInfo structure.
+        ///
+        /// This corresponds to [`i2d_PKCS8PrivateKey_bio`].
+        ///
+        /// [`i2d_PKCS8PrivateKey_bio`]: https://www.openssl.org/docs/man1.1.1/man3/i2d_PKCS8PrivateKey_bio.html
+        private_key_to_der_pkcs8,
+        /// Serializes the private key to a DER-encoded PKCS#8 EncryptedPrivateKeyInfo structure.
+        ///
+        /// This corresponds to [`i2d_PKCS8PrivateKey_bio`].
+        ///
+        /// [`i2d_PKCS8PrivateKey_bio`]: https://www.openssl.org/docs/man1.1.1/man3/i2d_PKCS8PrivateKey_bio.html
+        private_key_to_der_pkcs8_passphrase,
+        ffi::i2d_PKCS8PrivateKey_bio
+    }
 }
 
 impl<T> fmt::Debug for PKey<T> {
@@ -403,7 +424,7 @@ impl PKey<Private> {
     }
 
     /// Deserializes a DER-formatted PKCS#8 private key, using a callback to retrieve the password
-    /// if the key is encrpyted.
+    /// if the key is encrypted.
     ///
     /// The callback should copy the password into the provided buffer and return the number of
     /// bytes written.
@@ -565,6 +586,18 @@ mod tests {
         assert!(pub_key.windows(10).any(|s| s == b"PUBLIC KEY"));
     }
 
+    #[test]
+    fn test_der_pkcs8() {
+        let key = include_bytes!("../test/key.der");
+        let key = PKey::private_key_from_der(key).unwrap();
+
+        let priv_key = key.private_key_to_der_pkcs8().unwrap();
+
+        // Check that this has the correct PKCS#8 version number and algorithm.
+        assert_eq!(hex::encode(&priv_key[4..=6]), "020100"); // Version 0
+        assert_eq!(hex::encode(&priv_key[9..=19]), "06092a864886f70d010101"); // Algorithm RSA/PKCS#1
+    }
+
     #[test]
     fn test_rsa_accessor() {
         let rsa = Rsa::generate(2048).unwrap();
